"""Test the MetaboliteCorrectorFactory."""

import pytest
import isocor as hrcor


def test_MetaboliteCorrectorFactory(data_iso):
    """Test that the Factory can be instanciated safely."""
    # Low resolution
    x = hrcor.MetaboliteCorrectorFactory(
        "C3H7O6P", '13C', data_isotopes=data_iso)
    assert isinstance(x, hrcor.mscorrectors.LowResMetaboliteCorrector)
    # High resolution
    x = hrcor.MetaboliteCorrectorFactory("C3H7O6P", '13C', data_isotopes=data_iso,
                                         resolution=1e4, mz_of_resolution=400)
    assert isinstance(x, hrcor.mscorrectors.HighResMetaboliteCorrector)


@pytest.mark.parametrize("bad_dataiso", [{'C': {}},
                                         {'C': {'masses': [], 'abundance': []}},
                                         {'C': {'mass': [12, 13],
                                                'abundance': [1.]}},
                                         {'C': {'mass': [12, 13],
                                                'abundance': [1., 1.]}},
                                         {'C': {'mass': [12, 14], 'abundance': [.5, .5]}}])
def test_isotopedata(bad_dataiso):
    """Tests expected to fail because of a badly formatted data_isotopes."""
    with pytest.raises(ValueError):
        hrcor.MetaboliteCorrectorFactory("C", '13C', data_isotopes=bad_dataiso)

def test_typo_factory(data_iso):
    """Test a typo in the name of one of the parameters (Factory)."""
    with pytest.raises(ValueError):
        hrcor.MetaboliteCorrectorFactory("C3PO", "13C",
                                         wrong_data_isotopes_parameter=data_iso,
                                         correct_NA_tracer=False,
                                         derivative_formula=None,
                                         tracer_purity=None)

def test_typo_parameter(data_iso):
    """Test a typo in the name of one of the parameters (LowRes)."""
    with pytest.raises(TypeError):
        hrcor.LowResMetaboliteCorrector("C3PO", "13C",
                                        wrong_data_isotopes_parameter=data_iso,
                                        correct_NA_tracer=False,
                                        derivative_formula=None,
                                        tracer_purity=None)

@pytest.mark.parametrize("kwargs", [
    {"formula":"C", "tracer":'13C', "data_isotopes":None,
     "resolution":"42o", "mz_of_resolution":420},
    {"formula":"C", "tracer":'13C', "data_isotopes":None,
     "resolution":"420", "mz_of_resolution":"42o"}
])
def test_badinput(kwargs, data_iso):
    """Test a wrong type of parameter value (Factory)."""
    kwargs["data_isotopes"] = data_iso
    with pytest.raises(ValueError):
        hrcor.MetaboliteCorrectorFactory(**kwargs)



"""Test high-resolution correction matrix when no isotopic species are resolved from the tracer isotopologues.

In this situation, the contribution of all elements and isotopes should be removed,
hence the high-resolution and low-resolution correction matrices should be the same.

The high-resolution correction matrix calculated by IsoCor is compared to

1. a theoretical correction matrix (with algebraic equations provided in the code for manual check), and
2. the low resolution correction matrix.
"""

import numpy as np
import pytest
import isocor as hrcor

# Test the correction matrix by comparison with theoretical matrices expressed using explicit functions (here pXY represents natural
# abundance of isotope X of element Y), which can be easily verified. The examples provided are simple enough for manual verification,
# yet they represent the many situations that can be encountered in the real world. Each situation is detailed in the field "comment".


@pytest.mark.parametrize("data", [{"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 3 isotopes",
                                   "formula": "C2O2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p16O**2, 0.0, 0.0], [p17O*p16O*2, p16O**2, 0.0], [p17O*p17O+p18O*p16O*2, p17O*p16O*2, p16O**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2, 0.0, 0.0], [2*p1H*p2H, p1H**2, 0.0], [p2H**2, 2*p1H*p2H, p1H**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, with NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p1H**2*p12C**2, 0.0, 0.0], [2*p1H*p2H*p12C**2+2*p12C*p13C*p1H**2, p1H**2*p12C, 0.0], [p2H**2*p12C**2+2*p12C*p13C*p1H*p2H*2+p1H**2*p13C**2, p12C*p1H*p2H*2+p1H**2*p13C, p1H**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 2 elements with 2 isotopes",
                                   "formula": "C2H2N2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2*p14N**2, 0.0, 0.0], [2*p1H*p2H*p14N**2+2*p14N*p15N*p1H**2, p1H**2*p14N**2, 0.0], [p2H**2*p14N**2+2*p14N*p15N*p1H*p2H*2+p1H**2*p15N**2, 2*p1H*p2H*p14N**2+2*p14N*p15N*p1H**2, p1H**2*p14N**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, with purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2, p1H**2*pur_12C, p1H**2*pur_12C*pur_12C], [2*p1H*p2H, p1H**2*pur_13C+2*p1H*p2H*pur_12C, p1H**2*pur_12C*pur_13C*2+2*p1H*p2H*pur_12C*pur_12C], [p2H**2, 2*p1H*p2H*pur_13C+p2H**2*pur_12C, p2H**2*pur_12C*pur_12C+p1H**2*pur_13C*pur_13C+p1H*p2H*2*pur_13C*pur_12C*2]]",
                                   "tracer_purity": [0.1, 0.9]},
                                  {"comment": "13C tracer, with derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "C2",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2*p12C**2, 0.0, 0.0], [2*p1H*p2H*p12C**2+2*p12C*p13C*p1H**2, p1H**2*p12C**2, 0.0], [p2H**2*p12C**2+2*p12C*p13C*p1H*p2H*2+p1H**2*p13C**2, 2*p1H*p2H*p12C**2+2*p12C*p13C*p1H**2, p1H**2*p12C**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, with derivative, no NA of the tracer element, with purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "C2",
                                   "resolution": 1e4,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2*p12C**2, pur_12C*(p1H**2*p12C**2), p1H**2*p12C**2*pur_12C*pur_12C], [2*p1H*p2H*p12C**2 + 2*p12C*p13C*p1H**2, pur_13C*(p1H**2*p12C**2) + pur_12C*(2*p1H*p2H*p12C**2 + 2*p12C*p13C*p1H**2), pur_13C*(pur_12C*(p1H**2*p12C**2)) + pur_12C*(pur_13C*(p1H**2*p12C**2) + pur_12C*(2*p1H*p2H*p12C**2 + 2*p12C*p13C*p1H**2))], [p2H**2*p12C**2 + 2*p12C*p13C*p1H*p2H*2 + p1H**2*p13C**2, pur_13C*(2*p1H*p2H*p12C**2+2*p12C*p13C*p1H**2) + pur_12C*(p2H**2*p12C**2+2*p12C*p13C*p1H*p2H*2+p1H**2*p13C**2), pur_12C*(pur_13C*(2*p1H*p2H*p12C**2+2*p12C*p13C*p1H**2) + pur_12C*(p2H**2*p12C**2+2*p12C*p13C*p1H*p2H*2+p1H**2*p13C**2)) + pur_13C*(pur_13C*(p1H**2*p12C**2) + pur_12C*(2*p1H*p2H*p12C**2 + 2*p12C*p13C*p1H**2))]]",
                                   "tracer_purity": [0.1, 0.9]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, with purity of the tracer element, tracer with 3 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "CO",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p16O*p12C, 0.0], [p17O*p12C+p16O*p13C, p12C]]",
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "CO",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p12C, 0.0], [p13C, p12C]]",
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.0], [0.0, 1.0]]",
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"comment": "18O tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O2",
                                   "tracer": "18O",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]",
                                   "tracer_purity": [0., 0., 1.0]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O2",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]",
                                   "tracer_purity": [0., 1.0, 0.]},
                                  {"comment": "2H tracer, no derivative, NA of the tracer element, with purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "H2",
                                   "tracer": "2H",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p1H**2, p1H*pur1H, pur1H**2], [p1H*p2H*2, p1H*pur2H + p2H*pur1H, pur1H*pur2H*2], [p2H**2, p2H*pur2H, pur2H**2]]",
                                   "tracer_purity": [0.1, 0.9]},
                                  {"comment": "2H tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 0 elements",
                                   "formula": "H2",
                                   "tracer": "2H",
                                   "derivative_formula": "",
                                   "resolution": 1e2,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]",
                                   "tracer_purity": [0., 1.0]}])
def test_high_res_cor_matrix_unresolved(data, data_iso, usr_tolerance):
    """Correction of high resolution MS data.

    This test is designed to evaluate the correction of high-resolution MS data
    where none of the isotopic species are resolved from the tracer isotopologues,
    i.e. the high-resolution correction algorithm should provides the same correction
    matrix as the one computed using the low-resolution correction algorithm.
    """
    # perform correction
    metabolite = hrcor.HighResMetaboliteCorrector(data["formula"], data["tracer"],
                                                  data_isotopes=data_iso,
                                                  resolution_formula_code="orbitrap",
                                                  correct_NA_tracer=data["correct_NA_tracer"],
                                                  tracer_purity=data["tracer_purity"],
                                                  resolution=data["resolution"],
                                                  mz_of_resolution=data["mz_of_resolution"],
                                                  derivative_formula=data["derivative_formula"])
    correction_matrix = metabolite.compute_correction_matrix()
    # Test against expected matrix
    p1H = data_iso["H"]["abundance"][0]
    p2H = data_iso["H"]["abundance"][1]
    p12C = data_iso["C"]["abundance"][0]
    p13C = data_iso["C"]["abundance"][1]
    p14N = data_iso["N"]["abundance"][0]
    p15N = data_iso["N"]["abundance"][1]
    p16O = data_iso["O"]["abundance"][0]
    p17O = data_iso["O"]["abundance"][1]
    p18O = data_iso["O"]["abundance"][2]
    pur_12C = data["tracer_purity"][0]
    pur_13C = data["tracer_purity"][1]
    pur1H = data["tracer_purity"][0]
    pur2H = data["tracer_purity"][1]
    expected_matrix = eval(data["expected_matrix"])
    # Low resolution correction matrix
    LRmetabolite = hrcor.LowResMetaboliteCorrector(data["formula"], data["tracer"],
                                                   data_isotopes=data_iso,
                                                   correct_NA_tracer=data["correct_NA_tracer"],
                                                   tracer_purity=data["tracer_purity"],
                                                   derivative_formula=data["derivative_formula"])
    LRcorrection_matrix = LRmetabolite.compute_correction_matrix()
    # Test for equality
    np.testing.assert_allclose(
        correction_matrix, expected_matrix, rtol=usr_tolerance)
    np.testing.assert_allclose(
        correction_matrix, LRcorrection_matrix, rtol=usr_tolerance)



import tkinter as tk
from tkinter import ttk
from tkinter import scrolledtext
from tkinter import filedialog
from tkinter import messagebox
from isocor.ui.isocordb import EnvComputing
import logging
import pandas as pd
import isocor as hr
from pathlib import Path
import numpy as np
import webbrowser

UTF8_TABLE_SUBCRIPS_INT = {'0': '\u2070', '1': '\u00B9', '2': '\u00B2', '3': '\u00B3',
                           '4': '\u2074', '5': '\u2075', '6': '\u2076', '7': '\u2077', '8': '\u2078', '9': '\u2079'}


class TextHandler(logging.Handler):
    """This class allows you to log to a Tkinter Text or ScrolledText widget"""

    def __init__(self, text):
        # run the regular Handler __init__
        logging.Handler.__init__(self)
        # Store a reference to the Text it will log to
        self.text = text

    def emit(self, record):
        msg = self.format(record)

        def append():
            self.text.configure(state='normal')
            self.text.insert(tk.END, msg + '\n')
            self.text.configure(state='disabled')
            # Autoscroll to the bottom
            self.text.yview(tk.END)

        # This is necessary because we can't modify the Text from other threads
        self.text.after(0, append)

class AutoScrollbar(tk.Scrollbar):
    # a scrollbar that hides itself if it's not needed.  only
    # works if you use the grid geometry manager.
    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            # grid_remove is currently missing from Tkinter!
            self.tk.call("grid", "remove", self)
        else:
            self.grid()
        Scrollbar.set(self, lo, hi)
    def pack(self, **kw):
        pass
    def place(self, **kw):
        pass

class PurityTracerManager(tk.Canvas):
    def __init__(self, master=None, **kwargs):
        tk.Canvas.__init__(self, master, **kwargs)
        self._initFrameInWindows()
        self.tracer_purity = []

    def _initFrameInWindows(self):
        self.frame = ttk.Frame(self)
        self.create_window(0, 0, anchor="nw", window=self.frame)

    def changeEntries(self, df, tracer):
        row = 0
        self.tracer_purity = []
        # canvas clear all method
        self.delete("all")
        self._initFrameInWindows()
        purity = (df['subscriptName'] ==
                  tracer.iloc[0]['subscriptName']).astype(int)
        for entry in df.itertuples():
            purityentry = tk.StringVar()
            purityentry.set(str(purity[entry.Index]))
            label_entry = ttk.Label(
                self.frame, text=entry.subscriptName).grid(row=row, column=0)
            purity_entry = ttk.Entry(
                self.frame, textvariable=purityentry).grid(row=row, column=2)
            self.tracer_purity.append(purityentry)
            row += 1


class GUIinterface(ttk.Frame):
    """GUI interface for isocor in tk widget"""

    def __init__(self, master=None):
        super().__init__(master)
        self.pack()
        self.baseenv = EnvComputing()
        # check the database files exists in the default path,
        # otherwise copy them from the example folder
        self.baseenv.initializeDB()
        # isotope database should be loaded at init and not changed afterwards
        isotopesfile=Path(self.baseenv.db_path, "Isotopes.dat")
        try:
            self.baseenv.registerIsopotes(isotopesfile)
        except Exception as err:
            messagebox.showerror("Error", err)
            quit()

        self.addSubcriptingName()
        self.cleanListTracer()
        self.log_level = 'INFO'
        self.createWidgets()

    def addSubcriptingName(self):
        self.baseenv.dfIsotopes['subscriptName'] = self.baseenv.dfIsotopes['isotope'].apply(
            self.subscriptingInt) + self.baseenv.dfIsotopes['element']

    def subscriptingInt(self, myint):
        return ''.join([UTF8_TABLE_SUBCRIPS_INT[i] for i in str(myint)])

    def process(self):
        "process Callback"

        # clean GUI logs frame
        self.cleanLog()

        # get correction parameters
        tracer = self.baseenv.dfIsotopes[self.baseenv.dfIsotopes['subscriptName']
                                         == self.isotopictracerCBB.get()]['name'].values[0]
        correct_NA_tracer = self.chVarNatAbTracer.get()
        data_isotopes = self.baseenv.dictIsotopes
        # check critical parameters and cancel processing if errors
        try:
            tracer_purity = [float(i.get()) for i in self.purityManager.tracer_purity]
            if any(i < 0 for i in tracer_purity) or any(i > 1 for i in tracer_purity) or sum(tracer_purity) != 1:
                messagebox.showerror("Error", "Purity values should be within the range [0, 1], and their sum should be 1.")
                return
        except:
            messagebox.showerror("Error", "Purity values should be within the range [0, 1], and their sum should be 1.")
            return
        if self.chVarHR.get():
            resolution_formula_code = self.formulaEntered.get()
            try:
                resolution = float(self.varMass.get())
                if resolution <= 0:
                    messagebox.showerror("Error", "Resolution should be a positive number.")
                    return
            except:
                messagebox.showerror("Error", "Resolution should be a positive number.")
                return
            try:
                mz_of_resolution = float(self.varMZ.get())
                if mz_of_resolution <= 0:
                    messagebox.showerror("Error", "mz at which resolution is measured should be a positive number.")
                    return
            except:
                messagebox.showerror("Error", "mz at which resolution is measured should be a positive number.")
                return

        try:
            derivativesfile=Path(self.varDatabasePath.get(), "Derivatives.dat")
            self.baseenv.registerDerivativesDB(derivativesfile)
        except Exception as err:
            messagebox.showerror("Error", err)
            return
        try:
            metabolitesfile=Path(self.varDatabasePath.get(), "Metabolites.dat")
            self.baseenv.registerMetabolitesDB(metabolitesfile)
        except Exception as err:
            messagebox.showerror("Error", err)
            return
        try:
            input_file = self.varInputPath.get()
            self.baseenv.registerDatafile(input_file)
            fin_base = str(Path(input_file).stem)
        except Exception as err:
            messagebox.showerror("Error", err)
            return

        # add a filehandler to the logger (to redirect logs to a file)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s', "%Y-%m-%d %H:%M:%S")
        log_file = Path(self.varOutputPath.get()).joinpath(fin_base + '.log')
        file_handler = logging.FileHandler(str(log_file), mode='w+')
        file_handler.setLevel(self.log_level)
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

        # log general information on the process
        self.logger.info('------------------------------------------------')
        self.logger.info("Correction process")
        self.logger.info('------------------------------------------------')
        self.logger.info("   data files")
        self.logger.info("      data file: {}".format(input_file))
        self.logger.info("      derivatives database: {}".format(derivativesfile))
        self.logger.info("      metabolites database: {}".format(metabolitesfile))
        self.logger.info("   correction parameters")
        self.logger.info("      isotopic tracer: {}".format(tracer))
        self.logger.info("      correct natural abundance of the tracer element: {}".format(correct_NA_tracer))
        self.logger.info("      isotopic purity of the tracer: {}".format(tracer_purity))
        if self.chVarHR.get():
            self.logger.info("      mode: high-resolution")
            self.logger.info("         formula code: {}".format(resolution_formula_code))
            self.logger.info("         instrument resolution: {}".format(resolution))
            if resolution_formula_code != 'constant':
                self.logger.info("         at mz: {}".format(mz_of_resolution))
        else:
            self.logger.info("      mode: low-resolution")
        self.logger.info("   natural abundance of isotopes")
        self.logger.info("   {}".format(data_isotopes))
        self.logger.info("   IsoCor version: {}".format(hr.__version__))

        # initialize error dict
        errors = {'labels':[], 'measurements':[]}

        # construct correctors for all (metabolite, derivative)
        labels = self.baseenv.getLabelsList()
        self.logger.info('------------------------------------------------')
        self.logger.info('Constructing correctors for all (metabolite, derivative)...')
        self.logger.info('------------------------------------------------')
        dictMetabolites = {}
        for label in labels:
            try:
                self.logger.debug("constructing {}...".format(label))
                if self.chVarHR.get():
                    dictMetabolites[label] = hr.MetaboliteCorrectorFactory(
                            formula=self.baseenv.getMetaboliteFormula(label[0]), tracer=tracer, resolution=resolution, label=label[0],
                            data_isotopes=data_isotopes, mz_of_resolution=mz_of_resolution,
                            derivative_formula=self.baseenv.getDerivativeFormula(label[1]), tracer_purity=tracer_purity,
                            correct_NA_tracer=correct_NA_tracer, resolution_formula_code=resolution_formula_code)
                else:
                    dictMetabolites[label] = hr.MetaboliteCorrectorFactory(
                            formula=self.baseenv.getMetaboliteFormula(label[0]), tracer=tracer, label=label[0],
                            data_isotopes=data_isotopes,
                            derivative_formula=self.baseenv.getDerivativeFormula(label[1]), tracer_purity=tracer_purity,
                            correct_NA_tracer=correct_NA_tracer)
                self.logger.info("{} successfully constructed.".format(label))
            except Exception as err:
                dictMetabolites[label] = None
                errors['labels'] = errors['labels'] + [label]
                self.logger.error("cannot construct {}: {}".format(label, err))

        # correct measurements for naturally occuring isotopes
        # note: the correction matrix is constructed only once (at first correction of a (metabolite, derivative))
        self.logger.info('------------------------------------------------')
        self.logger.info('Correcting raw MS data...')
        self.logger.info('------------------------------------------------')
        df = pd.DataFrame()
        for label in labels:
            metabo = dictMetabolites[label]
            series, series_err = self.baseenv.getDataSerie(label)
            for s_err in series_err:
                errors['measurements'] = errors['measurements'] + ["{} - {}".format(s_err, label)]
                self.logger.error("{} - {}: Measurement vector is incomplete, some isotopologues are not provided.".format(s_err, label))
            for serie in series:
                if metabo:
                    try:
                         valuesCorrected = metabo.correct(serie[1])
                         self.logger.info("{} - {}: processed".format(serie[0], label))
                    except Exception as err:
                         valuesCorrected = ([np.nan]*len(serie[1]), [np.nan]*len(serie[1]), [np.nan]*len(serie[1]), np.nan)
                         self.logger.error("{} - {}: {}".format(serie[0], label, err))
                         errors['measurements'] = errors['measurements'] + ["{} - {}".format(serie[0], label)]
                else:
                    valuesCorrected = ([np.nan]*len(serie[1]), [np.nan]*len(serie[1]), [np.nan]*len(serie[1]), np.nan)
                    errors['measurements'] = errors['measurements'] + ["{} - {}".format(serie[0], label)]
                    self.logger.error("{} - {}: (metabolite, derivative) object could not be constructed.".format(serie[0], label))
                for i, line in enumerate(zip(*(serie[1], valuesCorrected[0], valuesCorrected[1], valuesCorrected[2], [valuesCorrected[3]]*len(valuesCorrected[0])))):
                    df = pd.concat((df, pd.DataFrame([line], index=pd.MultiIndex.from_tuples([[serie[0], label[0], label[1], i]], names=[
                        'sample', 'metabolite', 'derivative', 'isotopologue']), columns=['area', 'corrected_area', 'isotopologue_fraction', 'residuum', 'mean_enrichment'])))

        # save results
        out_file = Path(self.varOutputPath.get()).joinpath(fin_base + '_res.tsv')
        df.to_csv(str(out_file), sep='\t')

        # summary results for logs
        self.logger.info('------------------------------------------------')
        self.logger.info("Correction process summary")
        self.logger.info('------------------------------------------------')
        self.logger.info("   number of samples: {}".format(len(self.baseenv.getSamplesList())))
        self.logger.info("   number of (metabolite, derivative): {}".format(len(labels)))
        nb_errors = len(errors['labels']) + len(errors['measurements'])
        self.logger.info("   errors: {}".format(nb_errors))
        if nb_errors:
            self.logger.info("      {} errors during construction of (metabolite, derivative) correctors".format(len(errors['labels'])))
            self.logger.info("      {} errors during correction of measurements".format(len(errors['measurements'])))
            self.logger.info("      detailed information on errors are provided above.")

        # remove filehandler from the logger, and close the file
        self.logger.removeHandler(file_handler)
        file_handler.close()
        # log levels:
        #self.logger.debug('debug message')
        #self.logger.info('info message')
        #self.logger.warn('warn message')
        #self.logger.error('error message')
        #self.logger.critical('critical message')

    def cleanLog(self):
        self.logstream.config(state="normal")
        self.logstream.delete(1.0, tk.END)
        self.logstream.config(state="disabled")

    def cleanData(self):
        self.datatext.config(state="normal")
        self.datatext.delete(1.0, tk.END)
        self.datatext.config(state="disabled")

    def cleanListTracer(self):
        self.cleanDfIsotopes = self.baseenv.dfIsotopes[(self.baseenv.dfIsotopes.abundance > 0) & (
            self.baseenv.dfIsotopes.abundance < 1.)]
        dfMaxAbundance = self.cleanDfIsotopes.groupby(
            "element", as_index=False)["abundance"].max()
        self.cleanDfIsotopes = self.cleanDfIsotopes[~(self.cleanDfIsotopes['abundance'].isin(
            dfMaxAbundance['abundance']) & self.cleanDfIsotopes['element'].isin(dfMaxAbundance['element']))]

    def loadData(self):
        "load data Callback"
        # Using try in case user types in unknown file or closes without choosing a file.
        try:
            name = filedialog.askopenfilename(initialdir="C:/Users/Batman/Documents/Programming/tkinter/",
                                              filetypes=(
                                                  ("Data File", "*.tsv"), ("All Files", "*.*")),
                                              title="Choose a file."
                                              )
            self.cleanLog()
            self.cleanData()
            self.varInputPath.set(name)
            self.varOutputPath.set(Path(name).parent)
            with open(name, 'r') as UseFile:
                self.datatext.configure(state='normal')
                self.datatext.insert(tk.INSERT, UseFile.read())
                self.datatext.configure(state='disable')
        except:
            pass

    def outputDir(self):
        "gui output path"
        path = filedialog.askdirectory()
        self.varOutputPath.set(path)

    def databaseDir(self):
        "gui database path"
        path = filedialog.askdirectory()
        self.varDatabasePath.set(path)
        self.update_DBpath()

    def enableHR(self):
        if self.chVarHR.get():
            for child in self.highResFrame.winfo_children():
                child.configure(state='enable')
            self.formulaEntered.configure(state='readonly')
        else:
            for child in self.highResFrame.winfo_children():
                child.configure(state='disable')
        self.enableAtmz(None)

    def enableAtmz(self, event):
        if self.formulaEntered.get() == 'constant' or not self.chVarHR.get():
            self.mzEntry.configure(state='disable')
            self.mzlbl.configure(state='disable')
        else:
            self.mzEntry.configure(state='enable')
            self.mzlbl.configure(state='enable')

    def updatePurity(self, event):
        tracer = self.baseenv.dfIsotopes[self.baseenv.dfIsotopes['subscriptName']
                                         == self.isotopictracerCBB.get()]
        dfSymbol = self.baseenv.dfIsotopes[self.baseenv.dfIsotopes['element']
                                           == tracer.iloc[0]['element']]
        self.purityManager.changeEntries(df=dfSymbol, tracer=tracer)

    def updateLogLevel(self):
        if self.chVarVerboseLog.get():
            self.log_level = "DEBUG"
        else:
            self.log_level = "INFO"
        self.logger.setLevel(self.log_level)

    def enablePurity(self):
        pass

    def update_DBpath(self):
        self.baseenv.db_path = Path(self.varDatabasePath.get())

    def createWidgets(self):
        content = ttk.Frame(self, padding=(3, 3, 12, 12))
        dataFrame = ttk.Frame(content, padding=(3, 3, 12, 12))
        optionFrame = ttk.Frame(content, padding=(3, 3, 12, 12))

        self.TracerOptFrame = ttk.LabelFrame(
        optionFrame, text='Tracer correction options')

        tr_lab = ttk.Label(text="Isotopic purity of the tracer")
        purityLblFrame = ttk.LabelFrame(
            self.TracerOptFrame, labelwidget=tr_lab)
        self.purityManager = PurityTracerManager(
            purityLblFrame, width=200, height=100, highlightthickness=0)
        # previous code didn't work, trying to implement this example: http://effbot.org/zone/tkinter-autoscrollbar.htm
        # does not work yet
        self.scrollPurity = AutoScrollbar(purityLblFrame)
        purityLblFrame.update_idletasks()
        self.purityManager.config(scrollregion=self.purityManager.bbox("all"))

        tracer_list = list(self.cleanDfIsotopes['subscriptName'])
        self.isotopictracerEntered = tk.StringVar('')
        self.isotopictracerlbl = ttk.Label(
            optionFrame, text="Isotopic tracer")
        self.isotopictracerCBB = ttk.Combobox(
            optionFrame, textvariable=self.isotopictracerEntered, values=tracer_list, state="readonly")
        self.isotopictracerCBB.bind("<<ComboboxSelected>>", self.updatePurity)
        # default value: 13C
        try:
            default_tracer = tracer_list.index(u"\u00B9\u00B3\u0043")
        except:
            default_tracer = 0
        self.isotopictracerCBB.current(default_tracer)
        self.updatePurity(None)

        self.chVarHR = tk.IntVar()
        self.R1 = tk.Radiobutton(optionFrame, text="Low resolution", variable=self.chVarHR, value=False, command=self.enableHR)
        self.R2 = tk.Radiobutton(optionFrame, text="High resolution", variable=self.chVarHR, value=True, command=self.enableHR)

        self.highResFrame = ttk.LabelFrame(
            optionFrame, text='High resolution parameters')
        self.formulalbl = ttk.Label(self.highResFrame, text="Resolution formula",)
        self.formulaEntered = ttk.Combobox(
            self.highResFrame, values=self.baseenv.formulas_code, state="readonly")
        self.formulaEntered.bind("<<ComboboxSelected>>", self.enableAtmz)
        self.formulaEntered.current(0)
        self.varMass = tk.StringVar()
        self.varMZ = tk.StringVar()
        self.varMass.set("60000")
        self.varMZ.set('400.0')
        masslbl = ttk.Label(self.highResFrame, text="instrument resolution")
        self.massEntry = ttk.Entry(
            self.highResFrame, textvariable=self.varMass)
        self.mzlbl = ttk.Label(self.highResFrame, text="at m/z")
        self.mzEntry = ttk.Entry(self.highResFrame, textvariable=self.varMZ)

        self.chVarVerboseLog = tk.IntVar()
        self.chVarNatAbTracer = tk.IntVar()
        self.chVarPurityTracer = tk.IntVar()
        self.chVerboseLog = ttk.Checkbutton(
            content, text="Verbose logs", variable=self.chVarVerboseLog, command=self.updateLogLevel)
        self.processButon = ttk.Button(
            content, text=" Process ", command=self.process)
        self.chNatAbTracer = ttk.Checkbutton(
            self.TracerOptFrame, text="Correct natural abondance of the tracer element", variable=self.chVarNatAbTracer)
        self.varInputPath = tk.StringVar()
        self.varOutputPath = tk.StringVar()
        self.varOutputPath.set(self.baseenv.home)
        self.varDatabasePath = tk.StringVar()
        self.varDatabasePath.set(self.baseenv.default_db)
        self.inputDataEntry = ttk.Entry(
            dataFrame, textvariable=self.varInputPath, state='readonly')
        self.loadbutton = ttk.Button(
            dataFrame, text=" Load Data ", command=self.loadData)
        self.outputDataEntry = ttk.Entry(
            dataFrame, textvariable=self.varOutputPath, state='readonly')
        self.outputPathSubmit = ttk.Button(
            dataFrame, text=" Output Data Path ", command=self.outputDir)
        self.databaseEntry = ttk.Entry(
            dataFrame, textvariable=self.varDatabasePath, state='readonly')
        self.databasePathSubmit = ttk.Button(
            dataFrame, text=" Databases Path ", command=self.databaseDir)
        scrolH = 10
        self.datatext = scrolledtext.ScrolledText(
            dataFrame, width=40, height=scrolH, wrap=tk.WORD)
        self.datatext.configure(state='disable')
        self.logstream = scrolledtext.ScrolledText(
            content, height=scrolH, wrap=tk.WORD, state="disabled")

        for child in self.highResFrame.winfo_children():
            child.configure(state='disable')

        content.grid(column=0, row=0, sticky='NSWE')
        dataFrame.grid(column=0, row=0, sticky='NSWE')
        optionFrame.grid(column=1, row=0, sticky='NSW')
        self.isotopictracerlbl.grid(column=0, row=0, sticky='NW')
        self.isotopictracerCBB.grid(column=0, row=1, sticky='NW')
        self.R1.grid(column=0, row=2, sticky='NW')
        self.R2.grid(column=0, row=2, sticky='NE')
        self.TracerOptFrame.grid(column=0, row=4, sticky='NWE', )
        self.highResFrame.grid(column=0, row=3, sticky='NWE', )
        self.formulalbl.grid(column=0, row=0, sticky='NW')
        self.formulaEntered.grid(column=0, row=1, sticky='NW')
        masslbl.grid(column=0, row=2, sticky='NW')
        self.massEntry.grid(column=0, row=3, sticky='NW')
        self.mzlbl.grid(column=1, row=2, sticky='NW')
        self.mzEntry.grid(column=1, row=3, sticky='NW')
        purityLblFrame.grid(column=0, row=5, sticky='NW')
        self.purityManager.grid(column=0, row=1, sticky='NW')
        self.scrollPurity.grid(column=1, row=1, sticky='NS')
        self.chNatAbTracer.grid(column=0, row=0, sticky='NW')
        self.inputDataEntry.grid(column=0, row=0, sticky='NWE')
        self.loadbutton.grid(column=0, row=1, sticky='NWE')
        self.outputDataEntry.grid(column=0, row=3, sticky='NWE')
        self.outputPathSubmit.grid(column=0, row=4, sticky='NWE')
        self.databaseEntry.grid(column=0, row=5, sticky='NWE')
        self.databasePathSubmit.grid(column=0, row=6, sticky='NWE')
        self.datatext.grid(column=0, row=2, sticky='NSWE')
        self.processButon.grid(column=0, row=1, columnspan=2, sticky='NWE')
        self.logstream.grid(column=0, row=2, columnspan=2, sticky='NWE')
        self.chVerboseLog.grid(column=0, row=3, sticky='NW')

        # create texthandler and formatter to display logs
        self.scroll_handler = TextHandler(self.logstream)
        formatter = logging.Formatter(
            '%(levelname)s - %(message)s')
        self.scroll_handler.setFormatter(formatter)

        # create logger (should be root to catch 'mscorrectors' logger)
        self.logger = logging.getLogger()
        self.updateLogLevel()

        # add the text handler to logger
        self.logger.addHandler(self.scroll_handler)


def openDoc():
    webbrowser.open_new(r"https://isocor.readthedocs.io/en/latest/")

def openGit():
    webbrowser.open_new(r"https://github.com/MetaSys-LISBP/IsoCor/")

def start_gui():
    root = tk.Tk()
    root.resizable(width = False, height = False)
    menubar = tk.Menu(root)
    root.config(menu = menubar)
    filemenu = tk.Menu(menubar, tearoff=0)
    menubar.add_cascade(label="File", menu=filemenu)
    filemenu.add_command(label="Exit", command=root.quit)
    helpmenu = tk.Menu(menubar, tearoff=0)
    menubar.add_cascade(label="Help", menu=helpmenu)
    helpmenu.add_command(label = "IsoCor project", command=openGit)
    helpmenu.add_command(label = "Documentation", command=openDoc)
    app = GUIinterface(master=root)
    app.master.title("IsoCor {}".format(hr.__version__))
    app.mainloop()



import argparse
import isocor as hr
import isocor.ui.isocordb
import pandas as pd
import io
import logging
from pathlib import Path
import sys


def process(args):
    # create logger (should be root to catch all 'mscorrectors' loggers)
    logger = logging.getLogger()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s', "%Y-%m-%d %H:%M:%S")
    # sends logging output to sys.stderr
    strm_hdlr = logging.StreamHandler()
    strm_hdlr.setFormatter(formatter)
    logger.addHandler(strm_hdlr)

    if hasattr(args, 'verbose'):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # create environment
    baseenv = isocor.ui.isocordb.EnvComputing()
    if hasattr(args, 'I'):
        baseenv.registerIsopotes(Path(args.I))
    else:
        baseenv.registerIsopotes()
    if hasattr(args, 'D'):
        baseenv.registerDerivativesDB(Path(args.D))
    else:
        baseenv.registerDerivativesDB()
    if hasattr(args, 'M'):
        baseenv.registerMetabolitesDB(Path(args.M))
    else:
        baseenv.registerMetabolitesDB()
    baseenv.registerDatafile(Path(args.inputdata))

    # get correction parameters
    data_isotopes = baseenv.dictIsotopes
    tracer = args.tracer
    tracer_purity = getattr(args, 'tracer_purity', None)
    if tracer_purity:
        if any(i < 0 for i in tracer_purity) or any(i > 1 for i in tracer_purity) or sum(tracer_purity) != 1:
            raise ValueError(
                "Purity values ({}) should be within the range [0, 1], and their sum should be 1.".format(tracer_purity))
    correct_NA_tracer = True if hasattr(args, 'correct_NA_tracer') else False
    resolution = getattr(args, 'resolution', None)
    mz_of_resolution = getattr(args, 'mz_of_resolution', None)
    resolution_formula_code = getattr(args, 'resolution_formula_code', None)
    HRmode = resolution or mz_of_resolution or resolution_formula_code
    if HRmode:
        if not resolution:
            raise ValueError(
                "Applying correction to high-resolution data: 'resolution' should be provided.")
        if not mz_of_resolution:
            raise ValueError(
                "Applying correction to high-resolution data: 'mz_of_resolution' should be provided.")
        if not resolution_formula_code:
            raise ValueError(
                "Applying correction to high-resolution data: 'resolution_formula' should be provided.")
        if resolution <= 0:
            raise ValueError(
                "Resolution '{}' should be a positive number.".format(resolution))
        if mz_of_resolution <= 0:
            raise ValueError(
                "mz at which resolution is measured '{}' should be a positive number.".format(mz_of_resolution))

    # log general information on the process
    logger.info('------------------------------------------------')
    logger.info("Correction process")
    logger.info('------------------------------------------------')
    logger.info("   data files")
    logger.info("      data file: {}".format(args.inputdata))
    logger.info("      derivatives database: {}".format(
        getattr(args, 'D', 'Derivatives.dat')))
    logger.info("      metabolites database: {}".format(
        getattr(args, 'M', 'Metabolites.dat')))
    logger.info("      isotopes database: {}".format(
        getattr(args, 'I', 'Isotopes.dat')))
    logger.info("   correction parameters")
    logger.info("      isotopic tracer: {}".format(tracer))
    logger.info("      correct natural abundance of the tracer element: {}".format(
        correct_NA_tracer))
    logger.info("      isotopic purity of the tracer: {}".format(tracer_purity))
    if HRmode:
        logger.info("      mode: high-resolution")
        logger.info("         formula code: {}".format(
            resolution_formula_code))
        logger.info("         instrument resolution: {}".format(resolution))
        if resolution_formula_code != 'constant':
            logger.info("         at mz: {}".format(mz_of_resolution))
    else:
        logger.info("      mode: low-resolution")
    logger.info("   natural abundance of isotopes")
    logger.info("   {}".format(data_isotopes))
    logger.info("   IsoCor version: {}".format(hr.__version__))

    # initialize error dict
    errors = {'labels': [], 'measurements': []}

    labels = baseenv.getLabelsList()
    logger.info('------------------------------------------------')
    logger.info('Constructing correctors for all (metabolite, derivative)...')
    logger.info('------------------------------------------------')
    dictMetabolites = {}
    for label in labels:
        try:
            logger.debug("constructing {}...".format(label))
            if HRmode:
                dictMetabolites[label] = hr.MetaboliteCorrectorFactory(
                    formula=baseenv.getMetaboliteFormula(label[0]), tracer=tracer, resolution=resolution, label=label[0],
                    data_isotopes=data_isotopes, mz_of_resolution=mz_of_resolution,
                    derivative_formula=baseenv.getDerivativeFormula(label[1]), tracer_purity=tracer_purity,
                    correct_NA_tracer=correct_NA_tracer, resolution_formula_code=resolution_formula_code)
            else:
                dictMetabolites[label] = hr.MetaboliteCorrectorFactory(
                    formula=baseenv.getMetaboliteFormula(label[0]), tracer=tracer, label=label[0],
                    data_isotopes=data_isotopes,
                    derivative_formula=baseenv.getDerivativeFormula(label[1]), tracer_purity=tracer_purity,
                    correct_NA_tracer=correct_NA_tracer)
            logger.info("{} successfully constructed.".format(label))
        except Exception as err:
            dictMetabolites[label] = None
            errors['labels'] = errors['labels'] + [label]
            logger.error("cannot construct {}: {}".format(label, err))

    logger.info('------------------------------------------------')
    logger.info('Correcting raw MS data...')
    logger.info('------------------------------------------------')
    df = pd.DataFrame()
    for label in labels:
        metabo = dictMetabolites[label]
        series, series_err = baseenv.getDataSerie(label)
        for s_err in series_err:
            errors['measurements'] = errors['measurements'] + ["{} - {}".format(s_err, label)]
            logger.error("{} - {}: Measurement vector is incomplete, some isotopologues are not provided.".format(s_err, label))
        for serie in series:
            if metabo:
                try:
                    valuesCorrected = metabo.correct(serie[1])
                    logger.info("{} - {}: processed".format(serie[0], label))
                except Exception as err:
                    valuesCorrected = ([pd.np.nan]*len(serie[1]), [pd.np.nan]
                                       * len(serie[1]), [pd.np.nan]*len(serie[1]), pd.np.nan)
                    logger.error("{} - {}: {}".format(serie[0], label, err))
                    errors['measurements'] = errors['measurements'] + \
                        ["{} - {}".format(serie[0], label)]
            else:
                valuesCorrected = ([pd.np.nan]*len(serie[1]), [pd.np.nan]
                                   * len(serie[1]), [pd.np.nan]*len(serie[1]), pd.np.nan)
                errors['measurements'] = errors['measurements'] + \
                    ["{} - {}".format(serie[0], label)]
                logger.error(
                    "{} - {}: (metabolite, derivative) object could not be constructed.".format(serie[0], label))
            for i, line in enumerate(zip(*(serie[1], valuesCorrected[0], valuesCorrected[1], valuesCorrected[2], [valuesCorrected[3]]*len(valuesCorrected[0])))):
                df = pd.concat((df, pd.DataFrame([line], index=pd.MultiIndex.from_tuples([[serie[0], label[0], label[1], i]], names=[
                    'sample', 'metabolite', 'derivative', 'isotopologue']), columns=['area', 'corrected_area', 'isotopologue_fraction', 'residuum', 'mean_enrichment'])))

    # summary results for logs
    logger.info('------------------------------------------------')
    logger.info("Correction process summary")
    logger.info('------------------------------------------------')
    logger.info("   number of samples: {}".format(
        len(baseenv.getSamplesList())))
    logger.info("   number of (metabolite, derivative): {}".format(len(labels)))
    nb_errors = len(errors['labels']) + len(errors['measurements'])
    logger.info("   errors: {}".format(nb_errors))
    if nb_errors:
        logger.info("      {} errors during construction of (metabolite, derivative) correctors".format(
            len(errors['labels'])))
        logger.info("      {} errors during correction of measurements".format(
            len(errors['measurements'])))
        logger.info("      detailed information on errors are provided above.")

    output = io.StringIO()
    df.to_csv(output, sep='\t')
    output.seek(0)
    print(output.read())


def parseArgs():
    parser = argparse.ArgumentParser(argument_default=argparse.SUPPRESS,
                                     description='correction of MS data for naturally occurring isotopes')

    parser.add_argument("inputdata", help="measurements file to process")
    parser.add_argument("-M", type=str, help="path to metabolites database")
    parser.add_argument("-D", type=str, help="path to derivatives database")
    parser.add_argument("-I", type=str, help="path to isotopes database")

    parser.add_argument("-t", "--tracer", type=str, required=True,
                        help='the isotopic tracer (e.g. "13C")')
    parser.add_argument("-r", "--resolution", type=float,
                        help='HR only: resolution of the mass spectrometer (e.g. "1e4")')
    parser.add_argument("-m", "--mz_of_resolution", type=float,
                        help='HR only: mz at which resolution is given (e.g. "400")')
    parser.add_argument("-f", "--resolution_formula_code", type=str,
                        choices=hr.HighResMetaboliteCorrector.RES_FORMULAS, help="HR only: spectrometer formula code")
    parser.add_argument("-p", "--tracer_purity", type=lambda s: [float(item) for item in s.split(',')],
                        help="purity vector of the tracer")
    parser.add_argument("-n", "--correct_NA_tracer",
                        help="flag to correct tracer natural abundance", action='store_true')
    parser.add_argument("-v", "--verbose",
                        help="flag to enable verbose logs", action='store_true')
    return parser


def start_cli():
    parser = parseArgs()
    args = parser.parse_args()
    process(args)



"""*Correctors* classes of IsoCor for mass spectrometry data.

Metabolite *correctors* are metabolite-centered objects used to correct
mass spectrometry data for natural abundance and tracer purity.

Metabolite *correctors* can be instanciated directly or through their factory (recommended).
"""

import math
import functools
import itertools as it
import logging
from decimal import Decimal as D
import numpy as np
from scipy.optimize import fmin_l_bfgs_b
from isocor.base import LabelledChemical, InterfaceMSCorrector

logger = logging.getLogger(__name__)


class MetaboliteCorrectorFactory(object):
    """A Factory that returns the right *metabolite corrector* given the correction options.

    Args:
        formula (str): elemental formula of the metabolite moiety (e.g. "C3H7O6P")
        tracer (str): the isotopic tracer (e.g. "13C")
        label (str): metabolite abbreviation (e.g. "G3P")
        data_isotopes (dict): isotopic data with mass and abundance
            as in :py:attr:`~LabelledChemical.DEFAULT_ISODATA` (default values)
        derivative_formula (str): elemental formula of the derivative moiety (default:
            no derivative)
        tracer_purity (list): proportion of each isotope of the tracer element (metabolite moiety)
            The list must have the same length as the list of isotopes for the relevant
            isotope in :py:attr:`~data_isotopes`. Must be normalized to one.
            Default is perfect purity.
        correct_NA_tracer (bool): flag to correct tracer natural abundance or not.
            If set to True, tracer elements of the metabolite moiety will be
            corrected for the natural isotopic abundance of the tracer.
            Note that tracer elements from the derivative moiety will always be
            corrected at natural abundance (by definition of a derivative moiety).
            Default is False.
        resolution (int): resolution of the mass spectrometer (e.g. "1e4");
            default "None" (low-resolution).
        mz_of_resolution (int): m/z at which the resolution was measured (e.g. "400").
            This is **not** the m/z of the metabolite.
        resolution_formula_code (str): code for the resolution formula you wish to use to compute
            the correction limit among the presets: "orbitrap" (default), "ft-icr", and "constant".
            This formula depends on your mass spectrometer.

    Raises:
        ValueError: wrong input
        ImproperUsageError: incoherent input
    """
    def __new__(cls, formula, tracer, **kwargs):
        corrector = None
        # Gather common parameters and set default values
        label = kwargs.pop("label", None)
        data_isotopes = kwargs.pop("data_isotopes", None)
        derivative_formula = kwargs.pop("derivative_formula", None)
        tracer_purity = kwargs.pop("tracer_purity", None)
        correct_NA_tracer = kwargs.pop("correct_NA_tracer", False)
        # Gather up parameters used for specific correctors
        resolution = kwargs.pop("resolution", None)
        mz_of_resolution = kwargs.pop("mz_of_resolution", None)
        resolution_formula_code = kwargs.pop("resolution_formula_code", "orbitrap")
        # Choose a corrector
        if resolution is None and mz_of_resolution is None:
            logger.debug("MetaboliteCorrectorFactory chose to use a"
                         " LowResMetaboliteCorrector for %s.", formula)
            corrector = LowResMetaboliteCorrector(formula, tracer, label=label,
                                                  data_isotopes=data_isotopes,
                                                  derivative_formula=derivative_formula,
                                                  tracer_purity=tracer_purity,
                                                  correct_NA_tracer=correct_NA_tracer)
        elif resolution and mz_of_resolution:
            logger.debug("MetaboliteCorrectorFactory chose to use a"
                         " HighResMetaboliteCorrector for %s.", formula)
            try:
                corrector = HighResMetaboliteCorrector(formula, tracer, resolution, mz_of_resolution,
                                                       label=label,
                                                       data_isotopes=data_isotopes,
                                                       derivative_formula=derivative_formula,
                                                       tracer_purity=tracer_purity,
                                                       correct_NA_tracer=correct_NA_tracer,
                                                       resolution_formula_code=resolution_formula_code)
            except InterfaceMSCorrector.ImproperUsageError as reason:
                logger.warning("Improper usage of HighResMetaboliteCorrector "
                               "by MetaboliteCorrectorFactory."
                               " Falling back to LowResMetaboliteCorrector."
                               " Reason: %s", reason)
                corrector = LowResMetaboliteCorrector(formula, tracer, label=label,
                                                      data_isotopes=data_isotopes,
                                                      derivative_formula=derivative_formula,
                                                      tracer_purity=tracer_purity,
                                                      correct_NA_tracer=correct_NA_tracer)
        else:
            message = "MetaboliteCorrectorFactory was unable to select a" \
                      " correction strategy. Please check your inputs."
            logger.error(message)
            raise ValueError(message)
        # Warn user if a parameter some parameters were not utilized
        if kwargs:
            msg = "Unused parameters: {}. Maybe a typo?".format(str(kwargs))
            logger.error(msg)
            raise ValueError(msg)
        return corrector


class LowResMetaboliteCorrector(LabelledChemical, InterfaceMSCorrector):
    """Metabolite *corrector* for low-resolution mass-spectrometry data.

    At low-resolution, all isotopologues with the same nominal mass are
    considered to be measured together.

    Args:
        formula (str): elemental formula of the metabolite moiety (e.g. "C3H7O6P")
        tracer (str): the isotopic tracer (e.g. "13C")
        label (str): metabolite abbreviation (e.g. "G3P")
        data_isotopes (dict): isotopic data with mass and abundance
            as in :py:attr:`~LabelledChemical.DEFAULT_ISODATA`
        derivative_formula (str): elemental formula of the derivative moiety
        tracer_purity (list): proportion of each isotope of the tracer element (metabolite moiety)
            The list must have the same length as the list of isotopes for the relevant
            isotope in :py:attr:`~data_isotopes`. Must be normalized to one.
            Default is perfect purity.
        correct_NA_tracer (bool): flag to correct tracer natural abundance or not.
            If set to True, tracer elements of the metabolite moiety will be
            corrected for the natural isotopic abundance of the tracer.
            Note that tracer elements from the derivative moiety will always be
            corrected at natural abundance (by definition of a derivative moiety).
    """

    def __init__(self, formula, tracer, **kwargs):
        LabelledChemical.__init__(self, formula, tracer, **kwargs)
        InterfaceMSCorrector.__init__(self)
        # Log if direct instanciation
        if self.__class__.__name__ == LowResMetaboliteCorrector.__name__:
            self._log_on_init()

    def _log_on_init(self):
        """Log instantiation of a new metabolite corrector object."""
        key_debug_log = ["data_isotopes"]
        logme_info = []
        logme_debug = []
        for k, v in vars(self).items():
            if k not in key_debug_log:
                logme_info.append((k, v))
            else:
                logme_debug.append((k, v))
        logger.debug("New %s, %s: %s.", self.__class__.__name__,
                    self.label, logme_info)
        logger.debug("%s additionally uses: %s", self.label, logme_debug)

    def correct(self, measurement):
        """Return corrected measurement vector.

        Args:
            measurement (list): measured areas

        Returns:
            tuple:
                - corrected_area (array): Corrected area for each peak.
                - isotopologue_fraction (list): The abundance of each tracer
                  isotopologue (corrected area normalized to 1).
                - residuum
                - mean enrichment
        """
        logger.debug("New correction for %s with: measurement=%s.",
                     self.label, measurement)
        # Check length
        if len(measurement) != self.formula[self._tracer_el] + 1:
            raise ValueError("The length of the measured isotopic cluster ({}) is different"
                             " than the required number of measurements: {}"
                             " (i.e. N + 1, where N is the number of atoms that could"
                             " be traced)".format(len(measurement),
                                                  self.formula[self._tracer_el] + 1))
        # Perform the actual correction
        corrected_area, iso_fraction, residuum, enrichment = self._correct_with_bfgs(
            measurement)
        logger.debug(
            "Finished correction. Residuum (normalized to 1): %s", residuum)
        return corrected_area, iso_fraction, residuum, enrichment

    @staticmethod
    def _get_cost_function(mid, v_mes, mat_cor):
        """Cost function used for optimization.

        See also scipy.optimize.fmin_l_bfgs_b().

        Args:
            mid: isotopologue_fraction
            v_mes (array): measurement vector
            mat_cor (array): correction matrix

        Returns:
            float: (sum(v_mes - mat_cor * mid)^2, gradient)
        """
        x = v_mes - np.dot(mat_cor, mid)
        # calculate sum of square differences and gradient
        return (np.dot(x, x), np.dot(mat_cor.transpose(), x)*-2)

    def _correct_with_bfgs(self, measurement):
        """Perform the correction of the measurement vector using a L-BFGS-B algorithm.

        Args:
            measurement (list): measurement vector
        """
        # perform correction and calculate residuum
        v_mes = np.array(measurement).transpose()
        length_result = self.formula[self._tracer_el] + 1
        corrected_area, _, _ = fmin_l_bfgs_b(self._get_cost_function,
                                             np.zeros(length_result),
                                             fprime=None,
                                             approx_grad=0,
                                             args=(
                                                 v_mes, self.correction_matrix),
                                             factr=1000,
                                             pgtol=1e-10,
                                             bounds=[(0., float('inf'))] * length_result)
        resi = v_mes - np.dot(self.correction_matrix, corrected_area)
        # normalize mid and residuum
        sum_p = math.fsum(corrected_area)
        if sum_p != 0:
            isotopologue_fraction = [p/sum_p for p in corrected_area]
            enrichment = math.fsum(
                p*i for i, p in enumerate(isotopologue_fraction))/self.formula[self._tracer_el]
        else:
            isotopologue_fraction = [np.NaN for p in corrected_area]
            enrichment = np.NaN
        sum_m = math.fsum(measurement)
        if sum_m != 0:
            residuum = [v/sum_m for v in resi]
        else:
            residuum = [np.NaN for v in resi]
        return corrected_area, isotopologue_fraction, residuum, enrichment

    def get_mass_distribution_vector(self):
        """Get low resolution mass distribution vector (at natural abundance) for non-tracers elements.

        Based on the elemental compositions of both metabolite's and derivative's moieties.
        The element corresponding to the isotopic tracer is ignored.

        Calculation is based on convolution of isotopic vectors of individual elements (IsoCor,
        Millard et al., 2012), which is much faster than the combinatorial approach implemented for
        simulation at high resolution.

        Returns:
            list: mass distribution vector
        """
        result = [1.]  # mass are normalized to 1; also default value if no correction_formula
        for el, n in self.correction_formula.items():
            for _ in range(n):
                result = np.convolve(
                    result, self.data_isotopes[el]["abundance"])
        logger.debug("Done computing mass distribution vector for non-tracers "
                     "elements of %s (convolution method): %s", self.label, result)
        return list(result)

    def _correctionmatrix_convolution(self):
        """Return the correction matrix build with the convolution algorithm.

        Each column is constructed by iterative convolution of the correction vector.

        Returns:
            array: correction_matrix
        """
        correction_vector = self.get_mass_distribution_vector()
        n_isotopologues = self.formula[self._tracer_el] + 1
        # create correction matrix
        correction_matrix = np.zeros((n_isotopologues, n_isotopologues))
        # Peaks to keep after convolution
        mask = [n * self._idx_tracer for n in range(n_isotopologues)]
        # For each atom with the same element as the tracer
        for i in range(n_isotopologues):
            column = correction_vector
            # Correction of tracer purity
            for _ in range(i):
                column = np.convolve(column, self.tracer_purity)
            # Correct natural abundance of tracer element (if relevant)
            if self.correct_NA_tracer:
                for _ in range(n_isotopologues-i-1):
                    column = np.convolve(
                        column, self.data_isotopes[self._tracer_el]["abundance"])
            if len(column) < max(mask)+1:
                column += [0.]*(max(mask)-len(column)+1)
            column = [column[j] for j in mask]
            correction_matrix[:, i] = column
        logger.debug("Done computing correction matrix (convolution) for %s: %s",
                     self.label, correction_matrix)
        return correction_matrix

    def compute_correction_matrix(self):
        """Returns the correction matrix taking into account all parameters.

        Attention:
            Does not set :py:attr:`~correction_matrix` if called directly.

        Returns:
            array: the correction matrix
        """
        logger.debug("Computing correction matrix for %s...", self.label)
        return self._correctionmatrix_convolution()


class HighResMetaboliteCorrector(LowResMetaboliteCorrector):
    """Metabolite *corrector* for high-resolution mass-spectrometry data.

    At high-resolution, we consider that the measured peak **may not** encompass
    all isotopologues due to isotope natural abundance, depending on the operating
    resolution of the mass spectrometer at a given m/z.

    This is essentially the same correction pipeline as for low-resolution data,
    except that the calculation of the correction vector must be more precise and
    therefore requires a more complex (and more time-consumming) algorithm.

    Args:
        formula (str): elemental formula of the metabolite moiety (e.g. "C3H7O6P")
        tracer (str): the isotopic tracer (e.g. "13C")
        label (str): metabolite abbreviation (e.g. "G3P")
        data_isotopes (dict): isotopic data with mass and abundance
            as in :py:attr:`~LabelledChemical.DEFAULT_ISODATA`
        derivative_formula (str): elemental formula of the derivative moiety
        tracer_purity (list): proportion of each isotope of the tracer element (metabolite moiety)
            The list must have the same length as the list of isotopes for the relevant
            isotope in :py:attr:`~data_isotopes`. Must be normalized to one.
        correct_NA_tracer (bool): flag to correct tracer natural abundance or not.
            If set to True, tracer elements of the metabolite moiety will be
            corrected for the natural isotopic abundance of the tracer.
            Note that tracer elements from the derivative moiety will always be
            corrected at natural abundance (by definition of a derivative moiety).
        resolution (int): resolution of the mass spectrometer (e.g. "1e4")
        mz_of_resolution (int): m/z at which the resolution was measured (e.g. "400").
            This is **not** the m/z of the metabolite.
        resolution_formula_code (str): code for the resolution formula you wish to use to compute
            the correction limit among the presets: "orbitrap", "ft-icr", and "constant".
            This formula depends on your mass spectrometer.
        resolution_formula (func): (EXPERIMENTAL) if no code suits you, you
            can use this parameter to provide a function returning the resolution at the mz of the metabolite.
            Use the same format as in :attr:`~RES_FORMULAS`.
            :attr:`~resolution_formula` has precedence over :attr:`~resolution_formula_code`.
    """
    # TODO: resolution_formula is confusing: remove feat?
    # Registered resolution formulas to compute local resolution
    # parameters: molecular weight (mw), resolution (res) at mass-to-charge ratio (at_mz)
    RES_FORMULAS = {
        "orbitrap": lambda mw, res, at_mz: 1.66*mw**(3/2)/(res*math.sqrt(at_mz)),
        "ft-icr": lambda mw, res, at_mz: 1.66*mw**(3/2)/(res*math.sqrt(at_mz)),
        "constant": lambda mw, res, at_mz: 1.66*mw/res
    }

    def __init__(self, formula, tracer, resolution, mz_of_resolution, resolution_formula_code, **kwargs):
        LowResMetaboliteCorrector.__init__(self, formula, tracer, **kwargs)
        # Some checks on the inputs
        try:
            resolution = float(resolution)
            mz_of_resolution = float(mz_of_resolution)
        except ValueError as msg:
            raise ValueError("Parameters 'resolution' and 'mz_of_resolution' "
                             "should be numeric: {}".format(msg))
        if resolution <= 0.:
            raise ValueError(
                "'resolution' parameter should be >0 ({})".format(resolution))
        if mz_of_resolution <= 0.:
            raise ValueError(
                "'mz_of_resolution' parameter should be >0 ({})".format(mz_of_resolution))
        try:
            resolution_formula = kwargs.get("resolution_formula",
                                            self.RES_FORMULAS[resolution_formula_code])
        except KeyError:
            raise NotImplementedError("No resolution formula registered for code '{}'. "
                                      "Please provide the formula as resolution_formula"
                                      "parameter.".format(resolution_formula_code))

        self._correction_limit = resolution_formula(float(self.molecular_weight), resolution, mz_of_resolution)
        self.threshold_p = None if self.molecular_weight < 500 else 1e-10
        precision_machine = 10**3 * np.finfo(float).eps
        if self.correction_limit >= 0.5:
            raise self.ImproperUsageError("The correction limit is expected to be sufficent"
                                          " to distinguish peaks with a delta-mass of 1 amu"
                                          " ({}>0.5).".format(self.correction_limit))
        elif self.correction_limit < precision_machine:
            logger.warning("Correction limit is close to machine limits for floating point"
                           " operations. Correction limit reset to: %s", precision_machine)
            self._correction_limit = precision_machine
        # Log if direct instantiation
        if self.__class__.__name__ == HighResMetaboliteCorrector.__name__:
            self._log_on_init()

    @property
    def correction_limit(self):
        """Operative correction limit (in Da).

        Depends on the calibration resolution, the m/z at which the instrument
        was calibrated, the m/z of the meabolite and of course the formula that
        gives the relation between those values (that depends on the spectrometer).
        """
        return self._correction_limit

    # @staticmethod
    # def _count_isoblocks(n_isotopes, n_atoms):
    #     return int(math.factorial(n_isotopes+n_atoms-1) / math.factorial(n_atoms) / math.factorial(n_isotopes-1))

    @staticmethod
    def _count_isotopomers(isoblock):
        """Return the number of isotopomers for a given element/isoblock (permutations)."""
        dem = 1
        for x in set(isoblock):
            dem *= math.factorial(isoblock.count(x))
        return int(math.factorial(len(isoblock)) / dem)

    def _get_block_1(self, mass, n_atoms):
        return [(mass * n_atoms, 1.0)]

    def _get_block_2(self, masses, abundances, n_atoms):
        l_p = [1.]
        for i in range(n_atoms):
            l_p = np.convolve(l_p, abundances)
        res = [(masses[0]*(n_atoms-i) + masses[1]*i, l_p[i])
               for i in range(n_atoms+1)]
        return res

    def _get_block_n(self, masses, abundances, n_atoms, n_isotopes):
        all_res = []
        # Given a block of isotopes, fill its mass and proba
        all_isoblocks = it.combinations_with_replacement(
            range(n_isotopes), n_atoms)
        for isoblock in all_isoblocks:
            # Mass of this block
            block_mass = sum([masses[i] for i in isoblock])
            # Probability to see this block
            # NB: corrected by the number of isotopomere confounded in THIS block
            tmp_proba = [abundances[i] for i in isoblock]
            n_isotopomeres = self._count_isotopomers(isoblock)
            block_abundance = n_isotopomeres * \
                functools.reduce(lambda x, y: x*y, tmp_proba)
            all_res.append((block_mass, block_abundance))
        return all_res

    def _get_isotopic_blocks(self):
        """Return all combinations of isotopes by "block" of element."""
        data = {}  # mass and probability of each elemental block given its isotopic composition
        # Work element by element, to make elemental blocks (isotopes combinations)
        for element in self.correction_formula:
            n_isotopes = len(self.data_isotopes[element]["mass"])
            n_atoms = self.correction_formula[element]
            # TODO: check that the performances are improved with the different cases
            if n_isotopes == 1:
                # if element has only 1 isotope, probability is 1.0 by definition
                data[element] = self._get_block_1(
                    self.data_isotopes[element]["mass"][0], n_atoms)
            elif n_isotopes == 2:
                # if element has 2 isotope, use convolution method to calculate its mid
                data[element] = self._get_block_2(self.data_isotopes[element]["mass"],
                                                  self.data_isotopes[element]["abundance"],
                                                  n_atoms)
            else:
                # if element has 3 or more isotope, use full computation
                data[element] = self._get_block_n(self.data_isotopes[element]["mass"],
                                                  self.data_isotopes[element]["abundance"],
                                                  n_atoms, n_isotopes)
            # Keep only the blocks above probability threshold
            # NB: when the different elements are combined, peaks are always
            # multiplied by a number <1, so if any block is < threshold,
            # then the peak will be too
            assert all([x[1] >= 0 for x in data[element]]
                       ), "Unexpected negative probability."
            if self.threshold_p is not None:
                for peak in data[element]:
                    if peak[1] <= self.threshold_p:  # abundance
                        del peak
        return data

    def _combine_blocks(self, groups):
        """Combine blocks of isotopes together and compute the isotopic cluster."""
        def helper(X, Y):
            for x, y in it.product(X, Y):
                yield x[0] + y[0], x[1] * y[1]
        # generator over the data, by element
        g = (groups[element] for element in groups)
        iso_clust = dict()  # {mass: proba, ...}
        for mass, proba in functools.reduce(helper, g):
            if mass not in iso_clust:
                iso_clust[mass] = proba
            else:
                iso_clust[mass] += proba
        # Trim out the peaks below probability
        if self.threshold_p is not None:
            to_del = []
            for k in iso_clust:
                if iso_clust[k] <= self.threshold_p:
                    to_del.append(k)
            for k in to_del:
                del iso_clust[k]
        return iso_clust

    def get_isotopic_cluster(self):
        """Return the full isotopic cluster of the compound (for non-tracer elements).

        Args:
            threshold_p (float): Probability threshold under which a peak will be
                ignored (default: keep all peaks).

        Returns:
            dict: isotopologues masses mapped to their respective expected abundance
            (e.g. `{83.9959772546: 0.966808533640457, ...}`).
        """
        assert self.threshold_p is None or (0 <= self.threshold_p <= 1), \
            "Unexpected probability for 'threshold_p': {}".format(self.threshold_p)
        if self.correction_formula:
            groups = self._get_isotopic_blocks()
            iso_clust = self._combine_blocks(groups)
        else:
            iso_clust = None  # no atom, no isotopic cluster
        return iso_clust

    def get_tracershifted_peaks_between(self, mz_min, mz_max):
        """Return the mass of all the peaks that originate from the tracer.

        Those peaks are the 'main' peaks around which we will find other
        peaks due to the incorporation of natural isotopes of non-tracer elements
        (different from the most abundant isotope).

        Args:
            mz_min (float): minimum range
            mz_max (float): maximum range

        Returns:
            list: mass of all peaks that originate from the tracer
        """
        assert mz_max > mz_min, "Unexpected parameters: {} should be < to {}".format(
            mz_min, mz_max)
        n_peaks = math.floor((mz_max - mz_min)/self.mzshift_tracer) + 1
        id_mass = [mz_min + n * self.mzshift_tracer for n in range(n_peaks)]
        return id_mass

    def get_peaks_around(self, isotopic_cluster, masses):
        """Sum unresolved peaks of isotopic cluster according to the MS resolution.

        Args:
            isotopic_cluster (dict): isotopic cluster as returned by `:meth:~get_isotopic_cluster`
            masses (list): m/z around which peaks will be pooled

        Returns:
            list: abundances for each input mass, after sorting the mass and
            pooling nearby peaks within the resolution limit (:attr:`~correction_limit`)
        """
        # NB: masses are not garanteed to be in isotopic_cluster; if no peak is
        # found nearby, a 0.0 is returned for this mass.
        pooled = []
        for peak in sorted(masses):
            unresolved = [a for m, a in isotopic_cluster.items() if abs(
                m - peak) <= self.correction_limit]
            pooled.append(sum(unresolved))
        return pooled

    def get_mass_distribution_vector(self):
        """Get mass distribution vector (at natural abundancy) for non-tracers elements.

        Based on the elemental compositions of both metabolite's and derivative's moieties.

        Important:
            This methods needs to compute the full isotopic cluster,
            which is an expensive operation.

        Returns:
            list: mass distribution vector
        """
        iso_clust = self.get_isotopic_cluster()
        if iso_clust is None:  # all atoms are either tracer or ignored
            return [1.]     # same behavior as LowRes.
        minimum_mass = D(0)
        for element, n_elements in self.correction_formula.items():
            minimum_mass += n_elements * self.data_isotopes[element]["mass"][0]
        id_map = self.get_tracershifted_peaks_between(
            minimum_mass, max(iso_clust.keys()))
        mid = self.get_peaks_around(iso_clust, id_map)
        logger.debug("Done computing mass distribution vector for non-tracers "
                     "elements of %s (high-resolution method): %s", self.label, mid)
        return [float(x) for x in mid]

    def _correctionmatrix_combination(self):
        """Return the correction matrix constructed from the high-resolution correction vector (isotopic cluster).

        About the correction of tracer purity and natural abundance
        * Only atoms in *formula* are concerned; atoms of tracer's element in the derivative
          are already corrected in the correction_vector.
        * By default, we assume a perfect purity (see __init__).

        Returns:
            array: correction_matrix
        """
        n_tracers = self.formula[self._tracer_el]
        data_tracer = self.data_isotopes[self._tracer_el]
        n_tracer_isotopes = len(data_tracer["mass"])
        # Mass distribution vector without tracer
        isoclust_notracer = self.get_isotopic_cluster()
        if isoclust_notracer is not None:
            correction_vector = np.array(
                [(i, j) for i, j in isoclust_notracer.items()])
        main_peaks = self.get_tracershifted_peaks_between(self.molecular_weight,
                                                          self.molecular_weight + self.mzshift_tracer * n_tracers + D(0.5))
        # Prepare the Correction matrix
        n_isotopologues = n_tracers + 1
        correction_matrix = np.zeros((n_isotopologues, n_isotopologues))
        # For each measured peak, we compute the corresponding isotopic_cluster
        for n_traced_atoms in range(n_isotopologues):
            data = {}
            if isoclust_notracer is not None:
                data["cluster_notracer"] = correction_vector
            # Correct the tracer purity and natural abundance
            if n_traced_atoms:
                data["purity_tracer"] = self._get_block_n(data_tracer["mass"],
                                                          self.tracer_purity,
                                                          n_atoms=n_traced_atoms,
                                                          n_isotopes=n_tracer_isotopes)
            if n_tracers-n_traced_atoms:
                if self.correct_NA_tracer:
                    data["NA_tracer"] = self._get_block_n(data_tracer["mass"],
                                                          data_tracer["abundance"],
                                                          n_atoms=n_tracers-n_traced_atoms,
                                                          n_isotopes=n_tracer_isotopes)
                else:
                    data["NA_tracer"] = self._get_block_1(data_tracer["mass"][0],
                                                          n_tracers-n_traced_atoms)
            # Isotopic cluster with this combination of tracers
            iso_clust = self._combine_blocks(data)
            # Pool together unresolved peaks
            column = self.get_peaks_around(iso_clust, main_peaks)
            # Fill the corresponding column of the correction matrix
            correction_matrix[:, n_traced_atoms] = column
        logger.debug("Done computing correction matrix (isotopic-cluster method)"
                     " for %s: %s", self.label, correction_matrix)
        return correction_matrix

    def compute_correction_matrix(self):
        """Returns the correction matrix taking into account all parameters.

        Attention:
            Does not set :py:attr:`~correction_matrix` if called directly.

        Returns:
            array: the correction matrix
        """
        logger.debug("Computing correction matrix for %s...", self.label)
        n_tracer_isotopes = len(
            self.data_isotopes[self._tracer_el]["abundance"])
        assert n_tracer_isotopes > 1, "Unexpected number of isotopes for tracer ({}).".format(
            n_tracer_isotopes)
        if n_tracer_isotopes == 2:
            correction_matrix = self._correctionmatrix_convolution()
        else:
            correction_matrix = self._correctionmatrix_combination()
        return correction_matrix



"""IsoCor package initialisation.

We expose the Factory and the *metabolite corrector* classes at the package level
for conveniance.
"""

__version__ = '2.0.5'

from isocor.mscorrectors import MetaboliteCorrectorFactory
from isocor.mscorrectors import LowResMetaboliteCorrector, HighResMetaboliteCorrector



"""Test the calculation of low-resolution isotopic clusters (for metabolites with
different number of elements and isotopes) against a brute force implementation.
"""

import pytest
import numpy as np
import isocor as hrcor


def get_isoclust_bruteforce(formula, data_iso):
    """Return the low-resolution isotopic cluster of a compound using convolution."""
    result = [
        1.]  # mass are normalized to 1; also default value if no correction_formula
    for el, n in formula.items():
        for _ in range(n):
            result = np.convolve(result, data_iso[el]["abundance"])
    return list(result)


@pytest.mark.parametrize("formula", [{"C": 1, "N": 1},
                                     {"C": 1, "O": 1},
                                     {"C": 1, "N": 2},
                                     {"C": 1, "O": 2},
                                     {"C": 1, "N": 20, "H": 20, "O": 20, "P": 4},
                                     {"C": 20, "N": 20, "H": 20, "O": 20, "P": 4}])
def test_isoclust_against_bruteforce(formula, data_iso, usr_tolerance):
    """Check the low resolution isotopic clusters generation against a simple implementation."""
    str_formula = "".join(["{}{}".format(k, v) for k, v in formula.items()])
    # Get the clusters (without tracers)
    del formula["C"]
    ic_bruteforce = get_isoclust_bruteforce(formula, data_iso)
    metabolite = hrcor.LowResMetaboliteCorrector(str_formula, '13C',
                                                 data_isotopes=data_iso,
                                                 derivative_formula=None,
                                                 tracer_purity=None,
                                                 correct_NA_tracer=False)
    ic_optimized = metabolite.get_mass_distribution_vector()
    # Check that all isotopomeres are taken into account
    np.testing.assert_allclose(sum(ic_bruteforce), 1., rtol=usr_tolerance)
    np.testing.assert_allclose(sum(ic_optimized), 1., rtol=usr_tolerance)
    # Check that both methods have the same number of isotopologues
    assert len(ic_bruteforce) == len(ic_optimized)
    # Finally, check each abundance
    np.testing.assert_allclose(ic_bruteforce, ic_optimized, rtol=usr_tolerance)



"""Test the entire correction process at low- and high-resolution to cover all the situations that can be encountered.

This aims to verify that no error will be raised during normal usage.
This set of tests correct a measurement vector (set to unity, i.e. [1,1,1,1]
for a compound with 3 atoms of the tracer element) in all the possible combinations of the following situations:

* formulas: C1, C2, O1, O2, O3P3N3H3C3
* tracer elements: 13C, 17O, 18O
* derivative: H, H2, O, O2
* purity: 100%
* resolution (for high-resolution correction only): 10000, 100000, 1000000
* correct NA: True, False

Total number of tests: 240
"""

import numpy as np
import pytest
import isocor as hrcor


@pytest.mark.parametrize("data", [{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "O2",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "C2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "17O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0, 0.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O1",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O2",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "18O",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": True,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 10000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 100000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]},
{"formula": "O3P3N3H3C3",
"resolution_at_400": 1000000,
"derivative_formula": "H",
"tracer": "13C",
"correct_NA_tracer": False,
"tracer_purity": [0.0, 1.0]}])

def test_correction_process_misc(data, data_iso):
    """Tests the entire correction process at high resolution.

    Combination of all the following situations:
      - formulas: O3P3N3H3C3, C1, O1, O2, C2, O1
      - tracer elements: 13C (2 isotopes), 17O (3 isotopes), 18O (3 isotopes)
      - derivative: H, H2, O, O2
      - purity: 100%
      - resolution: 10000, 100000, 1000000
      - correct NA: True, False

    Total number of tests: 240
    """
    formula = data["formula"]
    derivative_formula = data["derivative_formula"]
    # Perform correction
    HRmetabolite = hrcor.HighResMetaboliteCorrector(formula, data["tracer"], data_isotopes=data_iso,
                                                    resolution=data["resolution_at_400"],
                                                    resolution_formula_code = "orbitrap",
                                                    mz_of_resolution=400,
                                                    derivative_formula=derivative_formula,
                                                    correct_NA_tracer=data["correct_NA_tracer"],
                                                    tracer_purity=data["tracer_purity"])
    LRmetabolite = hrcor.LowResMetaboliteCorrector(formula, data["tracer"], data_isotopes=data_iso,
                                                   correct_NA_tracer=data["correct_NA_tracer"],
                                                   derivative_formula=derivative_formula,
                                                   tracer_purity=data["tracer_purity"])
    tracer_element = data["tracer"][2]
    v_measured = [1] * (HRmetabolite.formula[tracer_element] + 1)
    _, v_correctedHR, _, _ = HRmetabolite.correct(v_measured)
    _, v_correctedLR, _, _ = LRmetabolite.correct(v_measured)



"""Configuration variables for the tests

Shared pre-defined parameters
The return value of fixture function will be available as a predefined
parameter for all test functions. The test's parameter name must be the same as
the fixture function's name.
"""

import pytest
from decimal import Decimal as D
import numpy as np


@pytest.fixture
def data_iso():
    """Typical isotopic data object."""
    return {"C": {"abundance": [0.7, 0.3],
                  "mass": [D('12.0'), D('13.0033548378')]},
            "H": {"abundance": [0.8, 0.2],
                  "mass": [D('1.0078250321'), D('2.014101778')]},
            "N": {"abundance": [0.9, 0.1],
                  "mass": [D('14.0030740052'), D('15.0001088984')]},
            "P": {"abundance": [1.],
                  "mass": [D('30.97376151')]},
            "O": {"abundance": [0.6, 0.3, 0.1],
                  "mass": [D('15.9949146221'), D('16.9991315'), D('17.9991604')]}}

@pytest.fixture
def usr_tolerance():
    """Platform-dependent tolerance."""
    return 10**4*np.finfo(float).eps



import pandas as pd
from pathlib import Path
from os.path import expanduser
import isocor as hr
from decimal import Decimal
from distutils.dir_util import copy_tree
import shutil
import numpy as np
import pkg_resources


class EnvComputing(object):
    """Share methods for interfaces"""

    def __init__(self, home=expanduser('~')):
        self.formulas_code = list(hr.HighResMetaboliteCorrector.RES_FORMULAS)
        self.formulas_code = self._putfirst(self.formulas_code, 'orbitrap')
        # initialize paths
        self.home = Path(home)
        self.default_db = Path(self.home, 'isocordb')
        self.db_path = self.default_db
        self.example_db = pkg_resources.resource_filename('isocor', 'data/')

    def initializeDB(self):
        # if db files don't exist, copy the example folder
        if self.default_db.is_dir():
            for i in ["Isotopes.dat", "Metabolites.dat", "Derivatives.dat"]:
                if not Path(self.default_db, i).is_file():
                    shutil.copy(Path(self.example_db, i), self.default_db)
        else:
            copy_tree(str(self.example_db), str(self.default_db))

    def initializeEnv(self):
        self.home = self.home

    def _putfirst(self, mylist, myelement):
        """return mylist with myelement as first position if is in"""
        if myelement in mylist:
            mylist.pop(mylist.index(myelement))
            mylist.insert(0, myelement)
        return mylist

    def _makeIsotopesDict(self, df):
        gb = df.groupby('element')
        d = {}
        for g in gb.groups:
            dg = gb.get_group(g).sort_values(by='mass').to_dict('list')
            e = dg.pop('element')[0]
            d[e] = dg
        return d

    def _stripColNames(self, df):
        df.rename(columns=lambda x: x.strip())

    def _stripCol(self, df, listcolumns):
        for col in listcolumns:
            df[col] = df[col].str.strip()

    def registerIsopotes(self, isotopesfile=Path("Isotopes.dat")):
        if not isotopesfile.is_file():
            raise ValueError(
                "Isotopes database not found in:\n'{}'.".format(isotopesfile))
        with open(str(isotopesfile), 'r') as fp:  # str for compatibility with Python3.5
            self.dfIsotopes = pd.read_csv(fp)
        for i in ['element', 'mass', 'abundance']:
            if i not in self.dfIsotopes.columns:
                raise ValueError("Column '{}' not found in 'Isotopes.dat'.".format(i))
        # check data types to return an explicit error
        if self.dfIsotopes.empty:
            raise ValueError("'Isotopes.dat' is empty.")
        for i, item in enumerate(self.dfIsotopes['mass']):
            try:
                Decimal(item)
            except:
                raise ValueError("Error in 'Isotopes.dat' at line {}:\nmass={!r}".format(i+2, item))
        for i, item in enumerate(self.dfIsotopes['abundance']):
            try:
                np.float64(item)
            except:
                raise ValueError("Error in 'Isotopes.dat' at line {}:\nabundance={!r}".format(i+2, item))
        # reload file
        with open(str(isotopesfile), 'r') as fp:
            self.dfIsotopes = pd.read_csv(
                 fp, converters={'mass': Decimal, 'abundance': np.float64})
        self._stripColNames(self.dfIsotopes)
        self._stripCol(self.dfIsotopes, ['element', ])
        self.dictIsotopes = self._makeIsotopesDict(self.dfIsotopes)
        self.dfIsotopes['isotope'] = self.dfIsotopes.mass.apply(round)
        self.dfIsotopes['name'] = self.dfIsotopes['isotope'].map(
            str) + self.dfIsotopes['element']

    def registerDerivativesDB(self, derivativesfile=Path("Derivatives.dat")):
        if not derivativesfile.is_file():
            raise ValueError(
                "Derivatives database not found in:\n'{}'.".format(derivativesfile))
        with open(str(derivativesfile), 'r') as fp:
            self.dfDerivatives = pd.read_csv(fp, delimiter='\t')
        for i in ['name', 'formula']:
            if i not in self.dfDerivatives.columns:
                raise ValueError("Column '{}' not found in 'Derivatives.dat'.".format(i))
        self._stripColNames(self.dfDerivatives)
        self._stripCol(self.dfDerivatives, ['name', 'formula'])

    def registerMetabolitesDB(self, metabolitesfile=Path("Metabolites.dat")):
        if not metabolitesfile.is_file():
            raise ValueError(
                "Metabolites database not found in:\n'{}'.".format(metabolitesfile))
        with open(str(metabolitesfile), 'r') as fp:
            self.dfMetabolites = pd.read_csv(fp, delimiter='\t')
        for i in ['name', 'formula']:
            if i not in self.dfMetabolites.columns:
                raise ValueError("Column '{}' not found in 'Metabolites.dat'.".format(i))
        self._stripColNames(self.dfMetabolites)
        self._stripCol(self.dfMetabolites, ['name', 'formula'])

    def registerDatafile(self, datafile=Path("mydata.tsv")):
        if not Path(datafile).is_file():
            raise ValueError("No data file selected.")
        with open(str(datafile), 'r') as fp:
            self.dfDatafile = pd.read_csv(fp, delimiter='\t')
        for i in ['sample', 'metabolite', 'derivative', 'area', 'isotopologue']:
            if i not in self.dfDatafile.columns:
                raise ValueError("Column '{}' not found in the data file.".format(i))
        # check data types to return an explicit error
        for i, item in enumerate(self.dfDatafile['area']):
            try:
                np.float64(item)
            except:
                raise ValueError("Error in data file at line {}:\narea={!r}".format(i+2, item))
        for i, item in enumerate(self.dfDatafile['isotopologue']):
            try:
                int(item)
            except:
                raise ValueError("Error in data file at line {}:\nisotopologue={!r}".format(i+2, item))
        self.dfDatafile[['area']] = self.dfDatafile[['area']].astype(np.float64)
        self.dfDatafile[['isotopologue']] = self.dfDatafile[['isotopologue']].astype(int)

        self.dfDatafile['derivative'].fillna('', inplace=True)
        if self.dfDatafile.empty:
            raise ValueError("Data file is empty.")
        self._stripColNames(self.dfDatafile)
        self._stripCol(self.dfDatafile, ['sample', 'metabolite', 'derivative'])
        self._groupbyDatafile = self.dfDatafile.groupby(
            by=['metabolite', 'derivative', 'sample'])

    def getLabelsList(self):
        return [tuple(i) for i in self.dfDatafile[['metabolite', 'derivative']].drop_duplicates().values]

    def getSamplesList(self):
        return [tuple(i) for i in self.dfDatafile[['sample']].drop_duplicates().values]

    def getMetaboliteFormula(self, name):
        try:
            return self.dfMetabolites[self.dfMetabolites['name'] == name]['formula'].values[0]
        except:
            raise ValueError(
                "No formula provided in 'Metabolites.dat' for metabolite '{}'.".format(name))

    def getDerivativeFormula(self, name):
        try:
            # derivative may be an empty str, otherwise should be in derivative database
            if name == '':
                return None
            else:
                return self.dfDerivatives[self.dfDerivatives['name'] == name]['formula'].values[0]
        except:
            raise ValueError(
                "No formula provided in 'Derivatives.dat' for derivative '{}'.".format(name))

    def returnLabelStr(self, tupleNames):
        if tupleNames[1]:
            return '-'.join(tupleNames)
        else:
            return tupleNames[0]

    def getDataSerie(self, tupleNames):
        l, l_err = [], []
        for i, j in self._groupbyDatafile:
            if i[:2] == tupleNames:
                try:
                    p=j.sort_values(by=['isotopologue'])
                    l.append([i[2], list(p.area.values)])
                except:
                    l_err.append(i[2])
        return l, l_err
